"""
The comparer module is responsible for analyzing the effects of 
changing a variable. It takes care of running the simulations, 
compiling the data, and producing graphs.
"""
import numpy as np
import shutil
import os
import os.path
import glob
import math
import ConfigParser as configparser
import warnings
import pkg_resources
#graphing
import matplotlib.pyplot as plt

from pykaryote.sim.environment import Environment
import pykaryote.utils.environment_draw as environment_draw
import pykaryote.utils.analysis as analysis
from pykaryote.sim.simulation import Simulation
from pykaryote.utils.globals import settings as gv

# ignore warnings about empty data files
warnings.filterwarnings('ignore', r'.*Empty input file.*')

# TODO: refactor metric settings into a config file
metrics_one_col =  ["complexity_avg", "irreducible_complexity_avg", 
"genome_length_avg", "fitness_avg", "1ancestors", "3ancestors", "10ancestors", 
"failed_builds", "gen_time"]
metrics_each_col = ["num_families_current_functional", 
"num_complexes_current_functional", "num_families_ever", "num_complexes_ever",
 "chemicals", "complexes", "percent_functional"]
metrics_logy = ['fitness','chemicals']
# metrics to plot bar graphs of the final values
# must be a list of values from metrics_one_col
metrics_to_plot_final_values = ['genome_length_avg', 'complexity_avg']

def get_config(filename):
    """Reads a comparison configuration file.

    Args:
        filename: The filename of comparison configuration file.

    Returns a tuple of (tuple of option names, list of tuples of values to use)
    
    If a line begins with a ``(``, it is simply ``eval``-ed. (options should 
        have quotes!)
    If the first line doesn't begin with a ``(``, then the text is surrounded 
        by quotes and put in a tuple.
    If subsequent lines don't begin with a ``(``, then they are simply 
        placed into tuples.
    """
    # TODO: rewrite cmp.cfg files to use json instead of eval
    lines = [l.strip() for l in open(filename)] #remove trailing newlines
    if lines[0][0]!='(':
        options = (lines[0],)
    else:
        options = eval(lines[0])
    values = []
    for l in lines[1:]:
         if len(l)>0:
            if l[0]!='(':
                values.append((eval(l),))
            else:
                values.append(eval(l))
    return options, values

def make_cfg(options, replacement_values, base_config, out_filename):
    """Create a sim config file by replacing options in base_config.

    Replaces options in base_config with values from replacement_values and
    saves the resulting simulation configuration file to out_filename.

    Used to create a sim.cfg config file from a cmp.cfg comparison config.
    """
    def replace_value(config, master_config, option, value):
        """Replaces a value in a configuration file.

        Checks to make sure the option exists either in the file,
        or in the master config file.
        """
        for section in config.sections():
            if config.has_option(section, option):
                config.set(section, option, value)
                return
        # option not found in config file
        # see if it is in master config file
        for section in master_config.sections():
            if master_config.has_option(section, option):
                try:
                    config.set(section, option, value)
                except configparser.NoSectionError:
                    config.add_section(section)
                    config.set(section, option, value)
                return
        # option not found in master config file
        raise ValueError('Unknown configuration option: {}'.format(option))

    # read in config file
    config = configparser.RawConfigParser()
    with open(base_config, 'r') as in_file:
        config.readfp(in_file, base_config)
    # read in master config file
    master_config = configparser.RawConfigParser()
    with pkg_resources.resource_stream('pykaryote',
                        os.path.join('configs', 'master.cfg')) as master_file:
        master_config.readfp(master_file)
    # replace values
    for option, value in zip(options, replacement_values):
        replace_value(config, master_config, option, value)
    # write output
    with open(out_filename, 'w') as out_file:
        config.write(out_file)

def get_metrics_files(folder, metric):
    """
    Gets the list of files associated with metric.
    
    *Args*:
        ``folder`` (str): the path to the "comparative-analyzer" folder. *Must* have a trailing '/'.
        ``metric`` (str): the name of the metric.
    
    *Returns* [str..]: the filenames that match.
    """
    files=[]
    i=1
    while os.path.isfile(os.path.join(folder, '%s-%s' % (metric, i))):
        files.append("%s-%i"%(metric,i))
        i+=1
    return files

class Comparer(object):
    """
    Runs a bunch of simulations with slight differences and aggregates the results.
    
    *Args*:
        ``sim_config`` (str) the template configuration file for the experiment
        
        ``cmp_config`` (str) the comparison configuration file for the experiment
        
        ``name`` (str): the name of the comparison; defaults to ``Experiment``
        
        ``data`` (str): the data directory; defaults to the current directory.
        
        ``identical_genomes`` (bool): whether all of the trials simulations 
        should start out with the same set of genomes (identical generation 0)
    """
    def __init__(self, sim_config, cmp_config, name="Experiment", 
                 data='.', clear=True, identical_genomes=True):
        self.vary_attrs,self.values = get_config(cmp_config)
        self.name = name
        self.identical_genomes = identical_genomes
        self.data = data
        self._sim_config = sim_config

        if clear:
            shutil.rmtree(os.path.join(self.data, name), ignore_errors=True)
        os.makedirs(os.path.join(self.data, name))

        #copy the comparison configuration into this comparison's own folder
        shutil.copy(cmp_config, os.path.join(self.data, name))

        #create configuration files for each varied option
        for x, vals in enumerate(self.values):
            out_filename = os.path.join(self.data, name, 'sim-%s.cfg' % x)
            make_cfg(self.vary_attrs, vals, sim_config, out_filename)
    
    def run(self, verbose=True):
        """
        
        *Args*:
            ``verbose``
        """
        #runs sims
        for i in range(len(self.values)):
            name = "sim-%s"%i
            if verbose:
                print "Simulating %s with %s=%s"%(name,self.vary_attrs,self.values[i])
            sim = Simulation(os.path.join(self.data, self.name,
                             'sim-%s.cfg' % i),
                            name=name, 
                            data=os.path.join(self.data, self.name),
                            comparison=True)
            if self.identical_genomes:
                if i==0: #this is the master simulation, all others will borrow its genomes
                    self.master_genomes = [o.genome for o in sim.organisms]
                else:
                    for o in range(len(sim.organisms)):
                        for b in range(len(sim.organisms[o].genome)):
                            sim.organisms[o].genome[b]=self.master_genomes[o][b]
            sim.run(verbose=verbose, log=True)
        #compile data
        agg = AnalyzerAggregator(self.data, self.name, self.values, 
                                 sim_config=self._sim_config)
        agg.compile_all()


class AnalyzerAggregator(object):
    """Aggregates data from multiple simulations into a comparative-analyzer
    directory.
    
    When a comparison is run, it creates an ``analyzer`` directory for each
    type of simulation. ``AnalyzerAggregator`` combines them by putting the
    columns next to each other. For example, combining three simulations, each
    with one column of data, produces a data file with three columns - one for
    each simulation.
    """
    def __init__(self, data_dir, comparison_name, cmp_values, sim_config):
        """
        Args:
            data_dir (str): path to a directory which holds comparisons.
            comparison_name (str): the name of this comparison
            cmp_values: a list of values, one for each type of simulation
                in the comparison.
            sim_config: path to the base simulaiton configuration file.
                Used to load settings for creating threshold graphs.
                (eg: gens_until_complexity_5)
        """
        self.data = data_dir
        self.name = comparison_name
        self.values = cmp_values
        # sim_config is used to load setting for which types of threshold metric
        # graphs to make
        self._sim_config = sim_config
        
    def compile_all(self):
        "Compiles all data files into a single ``comparative-analyzer`` dir."
        try:
            os.makedirs(os.path.join(self.data, self.name, 
                        'comparative-analyzer'))
        except:
            pass
        for m in metrics_one_col:
            self.compile_metric(m, 'one_col')
        for m in metrics_each_col:
            self.compile_metric(m, 'each_col')
        # compile threshold metrics
        if self._sim_config is not None:
            gv.init_globals(self._sim_config)
            for m in gv.metrics:
                name = 'gens_until_{}'.format(m)
                try:
                    # load previously calculated values from multiple runs
                    self.compile_metric('{}.stddev'.format(name), 'one_col')
                    self.compile_metric('{}.blanks'.format(name), 'one_col')
                    self.compile_metric(name, 'one_col')
                    self.compile_metric('num_runs', 'one_col')
                except IOError:
                    # single run
                    self._compile_threshold(name)
        # copy movement data
        self._compile_time_spent_in_gaussian()

    def _compile_time_spent_in_gaussian(self):
        """Copy ``time_spent_in_gaussian`` data to a comparative-analyzer dir.

        This data is used to look at movement. It tracks how long each organism
        spent in each chemical mound (gaussian).

        Subfolders are created inside of 'comparative-analyzer' for each
        simulation. Data files are copied to those subfolders.

        The data files themselves are three dimensional .np files. Each file
        is an array of 2d arrays, one for each run of a batch.
        """
        # copy data files
        for i in range(len(self.values)):
            analyzer_dir = os.path.join(self.data, self.name,
                                        'sim-{}'.format(i), 'analyzer')
            out_dir = os.path.join(self.data, self.name,
                                   'comparative-analyzer',
                                   'sim-{}'.format(i))
            try:
                os.makedirs(out_dir)
            except OSError:
                # directory already exists
                pass
            for data_file in glob.glob(os.path.join(
                                       analyzer_dir,
                                       'time_spent_in_gaussian_gen_*')):
                shutil.copy(data_file, out_dir)

    def _compile_threshold(self, metric):
        """Compiles threshold data for comparisons based off of a single run.

        Comparisons which are aggregates of multiple runs will already have
        this data generated for them.
        """
        out_path = os.path.join(self.data, self.name, 'comparative-analyzer',
                                metric)
        means = []
        stddevs = []
        blankss = []
        for i in range(len(self.values)):
            analyzer_dir = os.path.join(self.data, self.name, 
                                        'sim-{}'.format(i), 'analyzer')
            ann = analysis.Analyzer([analyzer_dir])
            mean, stddev, blanks = ann.quick_read(metric, mode='threshold')
            means.append(mean)
            stddevs.append(stddev)
            blankss.append(blanks)

        means = np.array(means)
        stddevs = np.array(stddevs)
        blankss = np.array(blankss)
        np.savetxt('{}.stddev'.format(out_path), stddevs.T)
        np.savetxt('{}.blanks'.format(out_path), blankss.T)
        np.savetxt(out_path, means.T)
        
    def compile_metric(self, metric, read_style):
        """Compile a single type of data file.
        
        *Args*:
            ``metric`` (str): the filename to read and write
            
            ``read_style`` (str): 
            
            * if ``one_col``, then the metric is assumed to have only one column.  
            * if ``each_col``, then each column is treated as its own metric, 
                and they are each written to "<metric_name>-<col#>"
            
        """
        datas = []
        if read_style=='one_col':
            #the lists contained by datas vary in length, including being [].
            for i in range(len(self.values)):
                datas.append(np.loadtxt(os.path.join(self.data, self.name, 
                                 'sim-%s' % i, 'analyzer', metric), 
                                dtype=float))
                if datas[-1].shape == (): #get at least one dimension.
                    datas[-1] = np.array([datas[-1]], dtype=float)
            num_rows = [val.shape[0] for val in datas] #for each val
            # TODO: use np.savetxt here
            f = open(os.path.join(self.data, self.name, 'comparative-analyzer', 
                     metric), 'w')
            for row in range(max(num_rows)):
                for val in range(len(datas)):
                    if row < num_rows[val]: #there's a value; write it.
                        f.write("%3.4f "%datas[val][row])
                    else: # -1 is used when the data file is empty
                        f.write("-1 ")
                f.write("\n")
            f.close()
        elif read_style=='each_col': 
            #the lists contained by datas vary in length, but are never []. the lists contained by those don't vary in length for now.
            for i in range(len(self.values)):
                datas.append(np.loadtxt(os.path.join(self.data, self.name, 
                                 'sim-%s' % i, 'analyzer', metric), 
                                dtype=float, ndmin=2))
            num_cols = [val.shape[1] for val in datas]
            num_rows = [val.shape[0] for val in datas]
            for col in range(max(num_cols)): #each column becomes a file
                f = open(os.path.join(self.data, self.name, 
                         'comparative-analyzer', '%s-%s' % (metric, 1 + col)),
                        'w')
                for row in range(max(num_rows)):
                    for val in range(len(datas)):
                        if row < num_rows[val] and col < num_cols[val]:
                            f.write("%3.4f "%datas[val][row][col])
                        else:
                            # -1 is used when there is no more data in the file
                            f.write('-1 ')
                    f.write('\n')
                f.close()
        else:
            raise Exception("unrecognized read_style: %s"%read_style)


class Grapher(object):
    """
    Takes a comparison's analyzer directory and provides methods for 
    plotting the data.
    
    *Args*:
        ``data_folder`` (str): a folder which contains a folder 
        named ``comparative-analyzer``, which has the necessary data.
    """
    # TODO: pass thresholds as a 'gens_until_complexity_avg.thresholds' file
        # so that way sim_config does not need to be passed to Grapher
    def __init__(self, data_folder, sim_config, cmp_config_filename=None,
                 verbose=True):
        gv.init_globals(sim_config)
        self.base = data_folder
        if self.base[-1]!='/': self.base+='/'
        if cmp_config_filename is None:
            cmp_config_filename = os.path.join(self.base, 'cmp.cfg')
        self.vary_attrs, self.values = get_config(cmp_config_filename)
        # load the number of simulations run for each varied attribute
        try:
            num_runs = np.loadtxt(os.path.join(self.base,
                                  "comparative-analyzer", 'num_runs'), 
                                  dtype=float, ndmin=1)
            if min(num_runs) == max(num_runs):
                self.num_runs = '{}'.format(int(num_runs[0]))
            else:
                self.num_runs = '{} to {}'.format(int(min(num_runs)),
                                                  int(max(num_runs)))
        except IOError:
            self.num_runs = '1'
        self._verbose = verbose

    def _report_status(self, string):
        """If verbose output is enabled, prints a string to stdout.
        """
        if self._verbose:
            print string
        
    def save_all(self):
        "Saves images of the plots of all of the metrics"
        try: 
            os.makedirs(os.path.join(self.base, 'graphs'))
        except:
            pass
        for m in metrics_one_col:
            plt.figure()
            self.plot(m)
            plt.savefig(os.path.join(self.base, 'graphs', '%s.png' % m))
        for m in metrics_each_col:
            for f in get_metrics_files(os.path.join(self.base, 
                                       'comparative-analyzer'), m):
                plt.figure()
                self.plot(f)
                plt.savefig(os.path.join(self.base, 'graphs', '%s.png' % f))
        for m in metrics_to_plot_final_values:
            plt.figure()
            self.plot_final_values(m)
            plt.savefig(os.path.join(self.base, 'graphs', 
                        '%s-final-values.png' % m))
        for metric, values in zip(gv.metrics, gv.thresholds):
            ann_dir = os.path.join(self.base, 'comparative-analyzer')
            self.plot_threshold(metric, values, save_dir=ann_dir)
        self.draw_environment(save_dir=os.path.join(self.base, 'graphs'))
        self.save_time_spent_in_gaussian()
        
    def plot(self, metric):
        """
        
        *Args*:
            ``metric`` (str): the name of the file to be plotted, 
            inside of the directory "comparative-analyzer". 
            Columns must correspond to values of vary_attr and 
            rows to generations.
        """
        self._report_status('plotting %s' % metric)
        logy = sum([m in metric for m in metrics_logy]) > 0
        try:
            data = np.loadtxt(os.path.join(self.base, "comparative-analyzer",
                        metric), dtype=float, ndmin=2) #row=generation
        except IOError:
            print 'no data found in', metric, 'a'

        plt.axes([.1,.1,.65,.8])
        if logy and not any(row[0]>0 for row in data): #hack to allow logy even when all values are zero
            for x in data[len(data)/2]:
                if x>0:
                    plt.plot([x])
                    break
            else:
                plt.plot([1e-1])
        f=(plt.semilogy if logy else plt.plot)
        for c in range(data.shape[1]): #each column
            amount_done= 1.0 - float(c) / data.shape[1]
            f(
                data[:,c],
                linewidth=amount_done*2+.5,
                color=(.3-.3*amount_done,1-amount_done,.3*amount_done),
                label=str(self.values[c]),
                linestyle=["solid","dashed","dashdot","dotted"][c % 4]
            )
        plt.title("%s over %s runs" % (metric, self.num_runs))
        plt.xlabel("generation")
        plt.ylabel(metric)
        plt.legend(title=self.vary_attrs, bbox_to_anchor=(1.01 ,1.), 
                   loc=2, borderaxespad=0.)

    def plot_final_values(self, metric):
        """Plots a bar graph of the final values of a metric.

        Each var_attr is a bar on the x axis.
        metric is the y axis.
        The value of each bar is the value of metric at the last generation.

        Input files contain 2d numpy arrays containing values for metric, where
        each row is a generation and each column corresponds to a value 
        of self.vary_attrs.

        For example:
            x-axis (bars) = varying config parameter
                (ex: cost of genome length)
            y-axis = final value for a metric after 1000 runs 
                (ex: average complexity)

        Args:
            metric (str): The metric to plot.
        """
        self._report_status("plotting final values of %s" % metric)
        try:
            data = np.genfromtxt(os.path.join(self.base, "comparative-analyzer",
                metric), dtype=float) #row=generation
        except IOError:
            print 'no data found in', metric
            return
        # ensure data is a 2d array
        if len(data.shape) < 2:
            if len(self.values) == 1:
                # there is only one column
                # this happens when only one value is changed in sim.cfg
                data.resize(data.shape[0], 1)
            else:
                # there is only one row
                # this happens when only one generation is run
                data.resize(1, data.shape[0])
        # check if the file is empty (no data to graph)
        if data.shape[1] == 0:
            print 'data file for metric', metric, 'is empty'
            return
        num_bars = data.shape[1] # number of columns
        num_generations = data.shape[0] # number of rows
        bar_x_cords = np.arange(num_bars) # x locations for the bars
        metrics_at_final_generation = data[-1] # heights of the bars
        width = 0.2
        bar_labels = self.values
        p1 = plt.bar(bar_x_cords, metrics_at_final_generation, width,
                     color='#175EC8')
        plt.ylabel(metric)
        plt.xlabel(self.vary_attrs)
        plt.title('%s after %d generations over %s runs' % (metric, 
                  num_generations, self.num_runs))
        plt.xticks(bar_x_cords+width/2.0, bar_labels)

    def plot_threshold(self, metric, thresholds, save_dir=None):
        """Plots a scatter plot of the number of generations it took a metric
        to pass a threshold value.

        These graphs are specified in the 'threshold-metrics' section of 
        a simulation's configuration file.
        """
        def new_plot(save_dir=save_dir, window_title=self.base):
            if save_dir is not None:
                plt.figure()

        def end_plot(name, save_dir=save_dir):
            if save_dir is not None:
                plt.savefig(os.path.join(self.base, 'graphs',
                            '{}.png'.format(name)))

        self._report_status("plotting threshold gens of %s" % metric)

        dir_path = os.path.join(self.base, 'comparative-analyzer')
        data_fname = 'gens_until_{}'.format(metric)
        stddev_fname = '{}.stddev'.format(data_fname)
        blanks_fname = '{}.blanks'.format(data_fname)
        data = np.loadtxt(os.path.join(dir_path, data_fname), ndmin=2)
        stddev = np.loadtxt(os.path.join(dir_path, stddev_fname), ndmin=2)
        blanks = np.loadtxt(os.path.join(dir_path, blanks_fname), ndmin=2)
        # calculate the fraction of runs which did not reach the threshold
        try:
            num_runs = np.loadtxt(os.path.join(self.base,
                                  "comparative-analyzer", 'num_runs'),
                                  dtype=float, ndmin=1)
        except IOError:
            num_runs = np.ones(blanks.shape[1])
        percent_blanks = np.empty(blanks.shape)
        for col_num, col in enumerate(blanks.T):
            for i in range(len(col)):
                percent_blanks[i, col_num] = col[i] / num_runs[col_num]

        # graph each threshold which was recorded for this metric
        for threshold_index, threshold in enumerate(thresholds):
            # plot blanks graphs
            new_plot()
            num_bars = data.shape[1]
            num_options = data.shape[0]
            bar_x_cords = np.arange(num_bars)
            percent_blanks_this_threshold = percent_blanks[threshold_index]
            width = 0.2
            bar_labels = self.values
            threshold = thresholds[threshold_index]
            plt.ylim(0.0, 1.0)
            p1 = plt.bar(bar_x_cords, percent_blanks_this_threshold, width,
                         color='#175EC8')
            plt.ylabel('percent runs')
            plt.xlabel(self.vary_attrs)
            plt.title('percent runs which did not reach '\
                      '{} {} over {} runs'.format(metric, threshold,
                                                 self.num_runs))
            plt.xticks(bar_x_cords + width / 2.0, bar_labels)
            plt.subplots_adjust(bottom=0.10)
            end_plot('percent_not_passed_{}_{}'.format(metric, threshold))

            # plot gens_until graphs
            new_plot()
            # when blank runs appear in the middle 80%, draw the bar red
            # this way data with blanks in it are drawn in red to indicate
            # that the average is wrong
            failed_threshold = (1.0 - gv.mean_to_aggregate) / 2.0
            bad_mask = blanks[threshold_index] >= (num_runs *
                                                   failed_threshold).astype(int)
            good_mask = ~bad_mask
            bad_gens_until_threshold = data[threshold_index][bad_mask]
            good_gens_until_threshold = data[threshold_index][good_mask]
            bad_stddev = stddev[threshold_index][bad_mask]
            good_stddev = stddev[threshold_index][good_mask]
            bad_bar_x_cords = bar_x_cords[bad_mask]
            good_bar_x_cords = bar_x_cords[good_mask]
            if len(good_bar_x_cords):
                good = plt.bar(good_bar_x_cords, good_gens_until_threshold,
                               width, yerr=good_stddev, color='#175EC8',
                               ecolor='#0A122A',
                               label='threshold always reached')
            if len(bad_bar_x_cords):
                bad = plt.bar(bad_bar_x_cords, bad_gens_until_threshold, width,
                              yerr=bad_stddev, color='#C60F0F',
                              ecolor='#2A0A0A', label='contains blanks')
            plt.ylabel('generations')
            plt.xlabel(self.vary_attrs)
            plt.title('generations until {} {} over {} runs'.format(metric, 
                      threshold, self.num_runs))
            plt.xticks(bar_x_cords + width / 2.0, bar_labels)
            plt.axis('auto')
            plt.ylim(ymin=0.0)
            plt.subplots_adjust(bottom=0.10)
            leg = plt.legend(loc='best', fancybox=True)
            leg.get_frame().set_alpha(0.5)
            end_plot('gens_until_{}_{}'.format(metric, threshold))

    def draw_environment(self, save_dir):
        """Draws the chemical environment and saves as an image.
        """
        env = Environment((gv.rows, gv.columns), log=False)
        environment_draw.draw_environment(env.grid, os.path.join(
                                        save_dir, 'environment.png'))

    def save_time_spent_in_gaussian(self):
        """Plots ``time_spent_in_gaussian`` data.

        Data should only be present if num_runs == 0.
        Data files are located in folders named
        ``comparative-analyzer/sim-x/``, where ``x`` is a number associated
        with each simulation.
        """
        save_dir=os.path.join(self.base, 'graphs')
        for sim_dir in glob.glob(os.path.join(self.base,
                                 'comparative-analyzer', 'sim-*')):
            ann = analysis.Analyzer([sim_dir])
            ann.plot_time_spent_in_gaussian()
            sim_num = os.path.split(sim_dir)[1][4:]
            plt.savefig(os.path.join(save_dir,
                        'time_spent_in_gaussians_sim_{}'.format(sim_num)))




# cython: embedsignature=True

cimport numpy as np

cdef class Complex:
    #cdef public list composition
    cdef public np.ndarray function
    cdef public family
    cdef public int length
    cdef public int is_functional 
    cdef public str name
    cdef public float binding_affinity

cdef class ComplexTracker:
    '''
    Tracks and manages all existing proteins
    '''
    cdef public dict complex_lookup
    cdef object complex_logfile
    cdef environment
    
    cdef Complex new_protein(self, str pname)
        
    cdef void log(self, Complex complex)
    cdef Complex new_complex(self, str complex, family)



#!/usr/bin/env python
import ez_setup
ez_setup.use_setuptools()
from setuptools import setup
from setuptools.extension import Extension
import glob
import sys
from os.path import join

try:
    import numpy
except ImportError:
    print 'Error: required package \'Numpy\' not found.'
    print '''Numpy is required in order to run this setup script. Install it \
    using your package manager or from numpy.org.

    While you're at it, make sure you have the following dependencies:
    \tpython-numpy
    \tpython-dev
    \topenmpi-bin
    '''
    print 'Numpy is required in order to run this setup script.'
    print 'Install it using your package manager or from numpy.org'
    print '\tex: sudo apt-get install python-numpy'
    print '\tor sudo pip install numpy'
    print 'On Linux, also ensure that the \'python-dev\' package is installed.'
    print '\n For detailed instructions, see documentation in the \'doc\' '\
          'directory'
    sys.exit(1)

def get_ext_modules():
    """Returns a list of extension modules to be build.

    If cython is installed, cython modules are cythonized and the 
    resulting C source code is returned.
    If cython is not installed, pre-cythonized C source code is returned.
    """
    numpy_dirs = numpy.get_include()
    # build cython files
    # try to compile .pyx files if cython is installed
    # otherwise fall back on compiling the pre-cythoned C code
    ext_modules = None
    try:
        from Cython.Build import cythonize
        print 'cythonizing .pyx files'
        ext_modules = cythonize(["pykaryote/sim/simulation.pyx",
                                "pykaryote/sim/environment.pyx",
                                "pykaryote/sim/families.pyx",
                                "pykaryote/sim/complexes.pyx",
                                "pykaryote/sim/complexes.pxd",
                                "pykaryote/sim/organism.pyx",
                                "pykaryote/sim/complexes.pxd",
                                "pykaryote/sim/genome.pyx"
                                ], )
    except ImportError:
        print 'cython not found, building pre-cythoned C code'
    except Exception:
        print 'cythoning failed, falling back on pre-cythoned C code'
    if ext_modules is None:
        ext_modules = [
            Extension("pykaryote.sim.simulation", 
                      ["pykaryote/sim/simulation.c"],
                      include_dirs = [".",numpy_dirs ]),
            Extension("pykaryote.sim.environment",
                      ["pykaryote/sim/environment.c"],
                      include_dirs = [".",numpy_dirs ]),
            Extension("pykaryote.sim.families", ["pykaryote/sim/families.c"],
                      include_dirs = [".",numpy_dirs ]),
            Extension("pykaryote.sim.complexes", ["pykaryote/sim/complexes.c"],
                      include_dirs = [".",numpy_dirs ]),
            Extension("pykaryote.sim.organism", ["pykaryote/sim/organism.c"],
                      include_dirs = [".",numpy_dirs ]),
            Extension("pykaryote.sim.genome", ["pykaryote/sim/genome.c"],
                      include_dirs = [".",numpy_dirs ]),
        ]
    return ext_modules

setup(
	name = "pykaryote",
	packages = ["pykaryote", "pykaryote.sim", "pykaryote.utils",
    'pykaryote.test', 'pykaryote.petri'],
	package_data={
    'pykaryote':['configs/*.cfg'],
    },
    include_package_data = True,
	version = "1.1.1",
	description = "Tools for modeling the evolution of biological complexity",
    long_description=open('README.txt').read(),
	author = "James Lamine, Ethan Van Andel, Peter Vandehaar",
	url = "https://pykaryote.readthedocs.org/en/latest/",
    install_requires = ['matplotlib', 'numpy', 'mpi4py', 'pillow', 'aggdraw'],
    provides = ['pykaryote'],
    scripts = glob.glob('bin/*'),
    classifiers = [
            "Programming Language :: Python",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7", 
            "Programming Language :: Cython",
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License (GPL)",  
            "Operating System :: OS Independent",
            "Topic :: Scientific/Engineering :: Artificial Life",
            "Topic :: Scientific/Engineering :: Physics"
            ],
            ext_modules = get_ext_modules()
    )



"""Contains functions for drawing the environment.
"""
from PIL import Image
from PIL import ImageColor
import aggdraw
import itertools
import os

CELL_SIZE = 8 # size of a grid cell in pixels
IMAGE_BOARDER = 2 # number of blanks cells around the image boarder

def draw_environment(grid, out_filename=None):
    """Draws the environment, either saving it to a file or returning an Image.
    """
    x_cells = grid.shape[0]
    y_cells = grid.shape[1]
    image_width = (x_cells + IMAGE_BOARDER * 2) * CELL_SIZE
    image_height = (y_cells + IMAGE_BOARDER * 2) * CELL_SIZE
    im = Image.new('RGB', (image_width, image_height), 'white')
    draw = aggdraw.Draw(im)
    draw.setantialias(True)
    draw_chemicals(draw, grid)
    #TODO: draw a legend
    draw_boarder(draw, x_cells, y_cells)
    draw.flush()
    if out_filename is not None:
        try:
            os.remove(out_filename)
        except Exception:
            pass
        im.save(out_filename)
    else:
        return im

def draw_chemicals(draw, grid):
    """Draws the distribution of chemicals in the environment.

    If ``out_filename`` is provided, the image is saved. Otherwise, a PIL 
    Image object is returned.

    Args:
        grid: A numpy array. Generally, this is an Environment.grid attribute.
    """
    x_cells = grid.shape[0]
    y_cells = grid.shape[1]
    num_chemicals = grid.shape[2]
    colors = make_chemical_colors(num_chemicals)
    for point in itertools.product(xrange(x_cells), xrange(y_cells)):
        for chemical, amount in enumerate(grid[point[0], point[1]]):
            brush = aggdraw.Brush(colors[chemical])
            #TODO: choose a better scale for percent. don't use hard coded 2.0
            scale_factor = amount
            #TODO: scale circle size by amount
            draw_circle(draw, point[0], point[1], scale_factor, brush=brush)

def draw_organisms(draw_object, grid_x, grid_y):
    """Draws the organisms on the environment grid.
    """

def draw_boarder(draw_object, grid_x, grid_y):
    """Draws a boarder around the environment grid.
    """
    x1 = y1 = IMAGE_BOARDER * CELL_SIZE - 1
    x2 = (IMAGE_BOARDER + grid_x) * CELL_SIZE + 1
    y2 = (IMAGE_BOARDER + grid_y) * CELL_SIZE + 1
    pen = aggdraw.Pen(ImageColor.getrgb('black'), width=1)
    draw_object.rectangle((x1, y1, x2, y2), pen, None)

def draw_circle(draw_object, cell_x, cell_y, size, pen=None, brush=None):
    """Draws a circle on ``draw_object`` in cell (cell_x, cell_y) with a
    diameter of ``size`` cells.
    """
    radius = size * CELL_SIZE / 2
    x1 = (cell_x + 0.5 + IMAGE_BOARDER) * CELL_SIZE - radius
    y1 = (cell_y + 0.5 + IMAGE_BOARDER) * CELL_SIZE - radius
    x2 = (cell_x + 0.5 + IMAGE_BOARDER) * CELL_SIZE + radius
    y2 = (cell_y + 0.5 + IMAGE_BOARDER) * CELL_SIZE + radius
    #assert(x2 - x1 == y2 - y1)
    draw_object.ellipse((x1, y1, x2, y2), pen, brush)

def make_chemical_colors(num_colors, sat=0.5, value=0.5, format='PIL'):
    """Returns a list of ``num_colors`` colors for coloring chemicals.

    sat and val are values between 0 and 1.
    """
    sat = int(sat * 100)
    value = int(value * 100)
    try:
        color_step = 360 / num_colors
    except ZeroDivisionError:
        return []
    rgb_colors = [ImageColor.getrgb('hsl({},{}%,{}%)'.format(hue, sat, value)) 
                    for hue in range(0, 360, color_step)]
    if format == 'matplotlib':
        rgb_colors = [(r/255.0, g/255.0, b/255.0) for r, g, b in rgb_colors]
    return rgb_colors



#!/usr/bin/env python
""" Globals loads settings from simulation configuration files.

Settings are loaded into an object called ``settings``
Typical usage::
    from pykaryote.utils.globals import settings as gv
    gv.init_globals(path_to_sim_cfg)
    x = gv.generation_limit

Configuration files have a weird, hackish syntax as a result of poor design
decisions early on. They are a mixture of INI and psudo-json. For more
information on simulation configuration file format, see the section of
that name in the reference documentation.

Ideally, config files would be converted to use pure JSON, but then backwards
compatibility with old config files would be lost.
"""
import ConfigParser as configparser
import pkg_resources
import json
import os.path
import re

def _parse_json_snippet(json_snippet):
    """Interprets a string as a snippet of json code.

    This allows json which is not wrapped as a full object to be loaded.
    """
    return json.loads('{{"val": {} }}'.format(json_snippet))['val']

def _read_config(config, dct=None):
    """Reads settings from a ConfigParser and returns a dict of option=value.

    Args:
        ``config`` (RawConfigParser): A RawConfigParser instance holding data
            from a sim.cfg file.
        ``dct`` (dict): A dictionary in which to store settings. If None,
            a dictionary is created.
    """
    if dct is not None:
        settings_dict = dct
    else:
        settings_dict = {}
    for section in config.sections():
        if section == 'runtime':
            # recorded_actions allows one to set arbitrary commands to run
            # on a given generation. This is used to study the effect of
            # changing the environment mid-simulation.
            # This code is unsafe because recorded actions are evaled.
            # TODO: re-write recorded actions to not use eval.
            recorded_actions = {}
            gens = [g for g in config.options(section)]
            for gen in gens:
                action_list = _parse_json_snippet(config.get(section, gen))
                recorded_actions[int(gen[3:])] = action_list
            if not 'recorded_actions' in settings_dict:
                settings_dict['recorded_actions'] = recorded_actions
            else:
                raise RuntimeError('Config file contains duplicate option: '
                                   'recorded_actions')
        for name, value in config.items(section):
            if not name in settings_dict:
                # coerce value to the appropriate type
                # try to load as a json object, list, number, or boolean
                try:
                    # Due to a poor design decision, the code used to use
                    # eval() to read config files. For safety, eval has been
                    # replaced with json, but in order to maintain backwards
                    # compatibility with old configuration files, a number of
                    # hacks are necessary.

                    # Parenthesis (python tuples) are treated as lists.
                    # Both 'True' and 'true' are interpreted as booleans
                    val = value.replace('(', '[').replace(')', ']').replace(
                            'True', 'true').replace('False', 'false')
                    # Floats without the leading zero are permitted
                    # Ex: both .5 and 0.5 are valid
                    val = re.sub(r'^\.(\d)', r'0.\1', val)
                    val = re.sub(r'([,\s\[\]])\.(\d)', r'\1 0.\2', val)
                    # parse the string as json
                    value = _parse_json_snippet(val)
                except ValueError:
                    # json parsing failed, treat value as a string
                    pass
                settings_dict[name] = value
            else:
                raise RuntimeError("Configuration file contains duplicate "
                                   "values for option: {}".format(name))
    return settings_dict


class SettingsMetaclass(type):
    """Creates the SettingsNamespace class.

    Creates and initializes attributes from the master configuration file.
    """
    def __new__(cls, name, bases, dct):
        config_parser = configparser.RawConfigParser()
        with pkg_resources.resource_stream('pykaryote',
                                           os.path.join('configs',
                                                        'master.cfg')) as f:
            config_parser.readfp(f)
        attrs = _read_config(config_parser, dct=dct)
        attrs['is_initialized'] = False
        return super(SettingsMetaclass, cls).__new__(cls, name, bases, attrs)


class SettingsNamespace(object):
    """Contains simulation configuration settings.

    Handles loading of configuration files. Settings default to those from
    ``master.cfg``. Uses a metaclass to create attributes to match the master
    config file.
    """

    __metaclass__ = SettingsMetaclass

    def __init__(self):
        self._update_calculations()

    def load(self, config_filename):
        """Loads data from a simulation configuration file.

        Data from the file overwrites the current settings. Missing data is
        not overwritten.
        Settings are type checked. If the types do not match those in
        'master.cfg', ValueError is thrown.
        """
        config_parser = configparser.RawConfigParser()
        with open(config_filename, 'r') as f:
            config_parser.readfp(f)
        settings_dict = _read_config(config_parser)
        for key, value in settings_dict.iteritems():
            if type(value) != type(getattr(self, key)):
                t1, t2 = type(value), type(getattr(self, key))
                if not (t1 in (float, int) and t2 in (float, int)):
                    # float and int mismatch is ok
                    raise ValueError('Type mismatch in attribute: {}.'.format(
                                     key)+' received: {}, expected: {}'.format(
                                     t1, t2))
            setattr(self, key, value)
        self._update_calculations()
        self.is_initialized = True

    def _update_calculations(self):
        """Loads settings. To be called whenever settings values are changed.
        """
        self.num_chemicals = sum(self.chem_list)
        self.min_codon = -(self.num_gather + self.num_move + self.num_protein)
        self.grange = range(-self.num_gather, 0)
        self.mrange = range(-(self.num_gather + self.num_move),
                            -self.num_gather)
        self.prange = range(self.min_codon, -(self.num_gather + self.num_move))

    def init_globals(self, config_file):
        """Alias for load(). Exists for backwards compatibility.
        """
        self.load(config_file)


def _initialize_settings():
    """Initializes ``settings`` object to default values from master.cfg file.

    To be called when the 'globals' module is loaded. This function uses
    meta-programming to create the 'SettingsNamespace' class from options
    in the config file, and the 'settings' instance is created. Defaults can
    be overwritten by calling settings.load(), which ensures that loaded
    settings are of the correct type.
    """
    global settings
    if settings is not None:
        raise RuntimeError('Settings have already been initialized to defaults.'
                           ' To override defaults, use settings.load()')
    settings = SettingsNamespace()


# load default settings from master.cfg the first time this module is imported
# TODO: ensure settings are only initialized the first time the module is
#  imported
settings = None
if settings is None:
    _initialize_settings()
